#ifndef __ALI_LIST_H__
#define __ALI_LIST_H__

typedef struct list_head {
    struct list_head *next, *prev;
}list_head_t;


/**
 * list_init_head - initialize a list head
 * @head: list head to initialize
 */
void list_init_head(list_head_t *head) ;


/**
 * list_add - add a new entry
 * @newitem: new entry to be added
 * @head: list head to add it after
 *
 * Insert a new entry after the specified head.
 * This is good for implementing stacks.
 */
void list_add(list_head_t *newitem, list_head_t *head);

/**
 * list_add_tail - add a new entry
 * @newitem: new entry to be added
 * @head: list head to add it before
 *
 * Insert a new entry before the specified head.
 * This is useful for implementing queues.
 */
void list_add_tail(list_head_t *newitem, list_head_t *head);

/**
 * list_del - deletes entry from list.
 * @item: the element to delete from the list.
 * Note: list_empty() on entry does not return true after this, the entry is
 * in an undefined state.
 */
void list_del(list_head_t *item) ;

/**
 * list_move - delete from one list and add as another's head
 * @item: the entry to move
 * @head: the head that will precede our entry
 */
void list_move(list_head_t *item, list_head_t *head) ;

/**
 * list_move_tail - delete from one list and add as another's tail
 * @item: the entry to move
 * @head: the head that will follow our entry
 */
void list_move_tail(list_head_t *item, list_head_t *head);

/**
 * list_items_swap - swap two items in list
 * @item1: the entry to swap
 * @item2: the entry to swap
 */
void list_items_swap(list_head_t *item1, list_head_t *item2);

/**
 * list_empty - tests whether a list is empty
 * @head: the list to test.
 */
int  list_empty(const list_head_t *head);


/*
 * Double linked lists with a single pointer list head.
 * Mostly useful for hash tables where the two pointer list head is
 * too wasteful.
 * You lose the ability to access the tail in O(1).
 */
#define LIST_HEAD_INIT(name) { &(name), &(name) }

#define LIST_HEAD(name) \
	struct list_head name = LIST_HEAD_INIT(name)

#define INIT_LIST_HEAD(ptr) do { \
	(ptr)->next = (ptr); (ptr)->prev = (ptr); \
} while (0)


#define _offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#define container_of(ptr, type, member) (                \
        (type *)( (char *)ptr - _offsetof(type,member)))

/**
* list_entry - get the struct for this entry
* @ptr:        the &struct list_head pointer.
* @type:       the type of the struct this is embedded in.
* @member:     the name of the list_struct within the struct.
*/
#define list_entry(ptr, type, member) \
        container_of(ptr, type, member)

/**
* list_for_each_entry_t  -  iterate over list of given type
* @pos:        the type * to use as a loop counter.
* @head:       the head for your list.
* @member:     the name of the list_struct within the struct.
* @type:       the type of the struct this is embedded in.
*/
#define list_for_each_entry_t(pos, head, member, type)	\
    for (pos = list_entry((head)->next, type, member);  \
         &pos->member != (head);    \
         pos = list_entry(pos->member.next, type, member))

#define list_for_each_entry_safe(pos, n, head, member, type)         \
    for (pos = list_entry((head)->next, type, member), \
         n = list_entry(pos->member.next, type, member);    \
         &pos->member != (head);                    \
         pos = n, n = list_entry(n->member.next, type, member))


/**
* list_for_each_entry_reverse_t - iterate backwards over list of given type.
* @pos:        the type * to use as a loop cursor.
* @head:       the head for your list.
* @member:     the name of the list_head within the struct.
* @type:       the type of the struct this is embedded in.
*/
#define list_for_each_entry_reverse_t(pos, head, member, type) \
    for (pos = list_entry((head)->prev, type, member);  \
         &pos->member != (head);    \
         pos = list_entry(pos->member.prev, type, member))

/**
* list_prepare_entry_t - prepare a pos entry for use in list_for_each_entry_continue()
* @pos:        the type * to use as a start point
* @head:       the head of the list
* @member:     the name of the list_head within the struct.
* @type:       the type of the struct this is embedded in.
*
* Prepares a pos entry for use as a start point in list_for_each_entry_continue().
*/
#define list_prepare_entry_t(pos, head, member, type) \
            ((pos) ? : list_entry(head, type, member))

/**
* list_for_each_entry_continue_t - iterate over list of given type
*                      continuing after existing point
* @pos:        the type * to use as a loop counter.
* @head:       the head for your list.
* @member:     the name of the list_struct within the struct.
* @type:       the type of the struct this is embedded in.
*/
#define list_for_each_entry_continue_t(pos, head, member, type) \
    for (pos = list_entry(pos->member.next, type, member);  \
         &pos->member != (head);    \
         pos = list_entry(pos->member.next, type, member))

/**
* list_for_each_entry_safe_t - iterate over list of given type safe against removal of list entry
* @pos:        the type * to use as a loop counter.
* @n:          another type * to use as temporary storage
* @head:       the head for your list.
* @member:     the name of the list_struct within the struct.
* @type:       the type of the struct this is embedded in.
*/
#define list_for_each_entry_safe_t(pos, n, head, member, type)	\
    for (pos = list_entry((head)->next, type, member),  \
        n = list_entry(pos->member.next, type, member); \
         &pos->member != (head);                    \
         pos = n, n = list_entry(n->member.next, type, member))
#endif
